@@ -1,27 +1,25 @@
-from __future__ import unicode_literals
-
 import os
-from threading import Lock
+from threading import RLock
 import warnings
 
 from .mmap_dict import mmap_key, MmapedDict
 
 
-class MutexValue(object):
+class MutexValue:
     """A float protected by a mutex."""
 
     _multiprocess = False
 
-    def __init__(self, typ, metric_name, name, labelnames, labelvalues, **kwargs):
+    def __init__(self, typ, metric_name, name, labelnames, labelvalues, help_text, **kwargs):
         self._value = 0.0
         self._exemplar = None
-        self._lock = Lock()
+        self._lock = RLock()
 
     def inc(self, amount):
         with self._lock:
             self._value += amount
 
-    def set(self, value):
+    def set(self, value, timestamp=None):
         with self._lock:
             self._value = value
 
@@ -52,15 +50,15 @@ def MultiProcessValue(process_identifier=os.getpid):
     # Use a single global lock when in multi-processing mode
     # as we presume this means there is no threading going on.
     # This avoids the need to also have mutexes in __MmapDict.
-    lock = Lock()
+    lock = RLock()
 
-    class MmapedValue(object):
+    class MmapedValue:
         """A float protected by a mutex backed by a per-process mmaped file."""
 
         _multiprocess = True
 
-        def __init__(self, typ, metric_name, name, labelnames, labelvalues, multiprocess_mode='', **kwargs):
-            self._params = typ, metric_name, name, labelnames, labelvalues, multiprocess_mode
+        def __init__(self, typ, metric_name, name, labelnames, labelvalues, help_text, multiprocess_mode='', **kwargs):
+            self._params = typ, metric_name, name, labelnames, labelvalues, help_text, multiprocess_mode
             # This deprecation warning can go away in a few releases when removing the compatibility
             if 'prometheus_multiproc_dir' in os.environ and 'PROMETHEUS_MULTIPROC_DIR' not in os.environ:
                 os.environ['PROMETHEUS_MULTIPROC_DIR'] = os.environ['prometheus_multiproc_dir']
@@ -71,7 +69,7 @@ def MultiProcessValue(process_identifier=os.getpid):
                 values.append(self)
 
         def __reset(self):
-            typ, metric_name, name, labelnames, labelvalues, multiprocess_mode = self._params
+            typ, metric_name, name, labelnames, labelvalues, help_text, multiprocess_mode = self._params
             if typ == 'gauge':
                 file_prefix = typ + '_' + multiprocess_mode
             else:
@@ -79,12 +77,12 @@ def MultiProcessValue(process_identifier=os.getpid):
             if file_prefix not in files:
                 filename = os.path.join(
                     os.environ.get('PROMETHEUS_MULTIPROC_DIR'),
-                    '{0}_{1}.db'.format(file_prefix, pid['value']))
+                    '{}_{}.db'.format(file_prefix, pid['value']))
 
                 files[file_prefix] = MmapedDict(filename)
             self._file = files[file_prefix]
-            self._key = mmap_key(metric_name, name, labelnames, labelvalues)
-            self._value = self._file.read_value(self._key)
+            self._key = mmap_key(metric_name, name, labelnames, labelvalues, help_text)
+            self._value, self._timestamp = self._file.read_value(self._key)
 
         def __check_for_pid_change(self):
             actual_pid = process_identifier()
@@ -101,13 +99,15 @@ def MultiProcessValue(process_identifier=os.getpid):
             with lock:
                 self.__check_for_pid_change()
                 self._value += amount
-                self._file.write_value(self._key, self._value)
+                self._timestamp = 0.0
+                self._file.write_value(self._key, self._value, self._timestamp)
 
-        def set(self, value):
+        def set(self, value, timestamp=None):
             with lock:
                 self.__check_for_pid_change()
                 self._value = value
-                self._file.write_value(self._key, self._value)
+                self._timestamp = timestamp or 0.0
+                self._file.write_value(self._key, self._value, self._timestamp)
 
         def set_exemplar(self, exemplar):
             # TODO: Implement exemplars for multiprocess mode.
